#include "config.h"
#include "ggap/gws.h"
#include "ggap/app.h"
#include "ggap/worksheet.h"
#include "zip/zip.h"
#include "zip/unzip.h"
#include <QtCore>
#include <stdlib.h>
#include <ctype.h>
#include <moo/macros.h>

using namespace ggap;

/**************************************************************************
 *
 * gws file is a zipped folder containing the following files:
 *  info.ini - meta info about the file
 *  workspace - saved workspace
 *  worksheet.xml - worksheet
 *
 */

static inline bool checkPlainText(QFile &file)
{
    file.seek(0);
    const QByteArray data = file.read(1024);
    for (int i = 0; i < data.size(); ++i)
        if (data[i] == 0 || (!isspace((uchar)data[i]) && iscntrl((uchar)data[i])))
            return false;
    return true;
}

bool ggap::guessFileType(const QString &filename, QString *type, QString &error)
{
    QFile file(filename);

    if (!file.open(QIODevice::ReadOnly))
    {
        error = file.errorString();
        return false;
    }

    char buf[4];
    qint64 n = file.read(buf, 4);

    if (n < 0)
    {
        error = file.errorString();
        return false;
    }

    if (n < 4 || memcmp(buf, "PK\x03\x04", 4) != 0)
    {
        if (!checkPlainText(file))
        {
            error = QApplication::translate("GwsReader", "bad file");
            return false;
        }
        else
        {
            *type = MimePlainText;
            return true;
        }
    }

    *type = MimeWorksheet;
    return true;
}

namespace ggap {

struct GwsInfo {
    QString workspace;
    QString worksheet;
    bool terminal;
    bool load(const QString &path, QString &error);
    void save(const QString &path);
};

#define INFO_INI "info.ini"
#define KEY_GWS_VERSION "gws/version"
#define CURRENT_VERSION "1.0"
#define KEY_GWS_WORKSHEET "gws/worksheet"
#define KEY_GWS_WORKSPACE "gws/workspace"
#define KEY_GWS_DATE_SAVED "gws/date-saved"
#define KEY_GWS_GGAP_VERSION "gws/ggap-version"
#define KEY_GWS_TERMINAL "gws/terminal"

bool GwsInfo::load(const QString &path, QString &error)
{
    QSettings s(path, QSettings::IniFormat);

    QString version = s.value(KEY_GWS_VERSION).toString();
    if (version != CURRENT_VERSION)
    {
        error = QApplication::translate("GwsReader", "invalid version '%1'").arg(version);
        return false;
    }

    workspace = s.value(KEY_GWS_WORKSPACE).toString();
    worksheet = s.value(KEY_GWS_WORKSHEET).toString();
    terminal = s.value(KEY_GWS_TERMINAL).toBool();
    return true;
}

void GwsInfo::save(const QString &path)
{
    QSettings s(path, QSettings::IniFormat);
    s.setValue(KEY_GWS_VERSION, CURRENT_VERSION);
    s.setValue(KEY_GWS_WORKSHEET, worksheet);
    if (terminal)
        s.setValue(KEY_GWS_TERMINAL, terminal);

    s.setValue(KEY_GWS_DATE_SAVED, QDateTime::currentDateTime().toString(Qt::ISODate));
    s.setValue(KEY_GWS_GGAP_VERSION, GGAP_VERSION);

    if (!workspace.isEmpty())
        s.setValue(KEY_GWS_WORKSPACE, workspace);
    s.sync();
}

} // namespace ggap

static void deleteDir(QDir &dir)
{
    QStringList files = dir.entryList();
    foreach (const QString &name, files)
        QFile::remove(dir.filePath(name));

    QFileInfo fi(dir.path());
    QDir(fi.path()).rmdir(fi.fileName());
}

/**************************************************************************
 *
 * GwsReader
 *
 */

struct ggap::GwsReaderPrivate {
    QString file;
    QString dir;
    QString workspace;
    QString worksheet;
    GwsReader *reader;
    bool terminal;

    GwsReaderPrivate(GwsReader *reader, const QString &file);
    ~GwsReaderPrivate();
    bool unpack(QString &error);
};

GwsReaderPrivate::GwsReaderPrivate(GwsReader *reader, const QString &file) :
    file(file), reader(reader), terminal(false)
{
}

GwsReaderPrivate::~GwsReaderPrivate()
{
}

bool GwsReaderPrivate::unpack(QString &error)
{
    m_return_val_if_fail(dir.isEmpty(), false);

    UnZip uz;
    UnZip::ErrorCode ec = uz.openArchive(file);
    if (ec != UnZip::Ok)
    {
        error = QApplication::translate("GwsReader", "openArchive failed: %1").arg(uz.formatError(ec));
        return false;
    }

    if (!uz.contains(INFO_INI))
    {
        error = QApplication::translate("GwsReader", "Invalid file");
        return false;
    }

    QDir d = gapApp->makeTempDir();
    if ((ec = uz.extractFile(INFO_INI, d)))
    {
        deleteDir(d);
        error = QApplication::translate("GwsReader", "extractFile failed: %1").arg(uz.formatError(ec));
        return false;
    }

    GwsInfo info;
    if (!info.load(d.filePath(INFO_INI), error))
    {
        deleteDir(d);
        return false;
    }

    if ((ec = uz.extractFile(info.worksheet, d)) ||
        (!info.workspace.isEmpty() && (ec = uz.extractFile(info.workspace, d))))
    {
        deleteDir(d);
        error = QApplication::translate("GwsReader", "extractFile failed for "
                                        "worksheet or workspace: %1").arg(uz.formatError(ec));
        return false;
    }

    if (!info.workspace.isEmpty())
        workspace = d.filePath(info.workspace);
    if (!info.worksheet.isEmpty())
        worksheet = d.filePath(info.worksheet);
    terminal = info.terminal;
    dir = d.absolutePath();
    return true;
}

GwsReader::GwsReader(const QString &file) :
    impl(new GwsReaderPrivate(this, file))
{
}

GwsReader::~GwsReader()
{
    delete impl;
}

bool GwsReader::unpack(QString &error)
{
    return impl->unpack(error);
}

QString GwsReader::workspace()
{
    return impl->workspace;
}

QString GwsReader::worksheet()
{
    return impl->worksheet;
}

bool GwsReader::terminal()
{
    return impl->terminal;
}


/**************************************************************************
 *
 * GwsWriter
 *
 */

struct ggap::GwsWriterPrivate {
    QString dest;
    QDir dir;
    GwsInfo info;
    bool closed;

    GwsWriterPrivate(const QString &dest, bool terminal) :
        dest(dest), closed(true)
    {
        info.terminal = terminal;
    }
};

GwsWriter::GwsWriter(const QString &dest, bool terminal) :
    impl(new GwsWriterPrivate(dest, terminal))
{
}

GwsWriter::~GwsWriter()
{
    close();
    delete impl;
}

bool GwsWriter::open(QString &)
{
    m_return_val_if_fail(impl->closed, false);
    impl->closed = false;
    impl->dir = gapApp->makeTempDir();
    return true;
}

QString GwsWriter::workspaceFilename()
{
    m_return_val_if_fail(!impl->closed, QString());
    QString workspace = "workspace";
    impl->info.workspace = workspace;
    return impl->dir.filePath(workspace);
}

bool GwsWriter::addWorksheet(const QByteArray &xml, QString &error)
{
    m_return_val_if_fail(!impl->closed, false);
    QString worksheet = "worksheet.xml";
    QString path = impl->dir.filePath(worksheet);
    QFile file(path);
    if (!file.open(QIODevice::WriteOnly))
    {
        error = file.errorString();
        return false;
    }
    if (file.write(xml) != xml.size() || !file.flush())
    {
        error = file.errorString();
        return false;
    }
    file.close();
    impl->info.worksheet = worksheet;
    return true;
}

bool GwsWriter::write(QString &error)
{
    m_return_val_if_fail(!impl->closed, false);
    impl->info.save(impl->dir.filePath(INFO_INI));

    QTemporaryFile *tmp_file = new QTemporaryFile(impl->dest + "XXXXXX");
    tmp_file->setAutoRemove(false);

    if (!tmp_file->open())
    {
        error = tmp_file->errorString();
        delete tmp_file;
        return false;
    }

    QString tmp_path = tmp_file->fileName();

    Zip zip;
    Zip::ErrorCode ec;
    if ((ec = zip.createArchive(tmp_file)) ||
        (ec = zip.addDirectoryContents(impl->dir.path(), Zip::Deflate3)) ||
        (ec = zip.closeArchive()))
    {
        error = zip.formatError(ec);
        return false;
    }

    QFile d(impl->dest);
    QFile f(tmp_path);
    if (d.exists() && !d.remove())
    {
        error = d.errorString();
        f.remove();
        return false;
    }
    if (!f.rename(impl->dest))
    {
        error = f.errorString();
        f.remove();
        return false;
    }

    return true;
}

void GwsWriter::close()
{
    if (impl->closed)
        return;

    deleteDir(impl->dir);
    impl->closed = true;
}
